/* ifile - intelligent mail filter for EXMH/MH
   Copyright (C) 1997  Jason Rennie <jr6b+@andrew.cmu.edu>
   
   irefile.c code Copyright (C) 1997 by Jason Rennie <jr6b+@andrew.cmu.edu>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program (see file 'COPYING'); if not, write to the Free
   Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA  02111-1307, USA.
   */

/*
  irefile.c

  C implementation of irefile.perl.

  Written by Jason Rennie, beginning Mar 30, 1997.

  This is a fairly straight translation from the perl code.  The only purpose
  at this point was to increase the efficiency of the program.	Some code was
  taken from the ifilter.c code.  The design of the code is very similar to
  that of the irefile.perl code.

  This program relies on the availability of the MH refile program.

  */

#include <sys/stat.h>	   // stat functions
#include <unistd.h>	   // stat functions
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <stdarg.h>       /* va_*() */
#include <assert.h>       /* assert() */
#include <unistd.h>       /* pipe() fork() close() dup2() execlp() */
#include <errno.h>	  /* required by Solaris 5.5.1/gcc-2.7.2 for errno variable */
#include <ctype.h>        /* tolower() */
#include <string.h>

#include "assoc_array_int.h"

#include "error.h"        /* ifile message and error reporting library */
#include "ifile.h"        /* standard ifile library */

/* Global variables used for storing information about documents, idata file */
assoc_array_int as_n,       /* number of total words in a category */
                as_docs,    /* number of documents in a category */
                as_age,     /* age of a word */
                as_nk,      /* instances of a word in a specific
			       category */
                as_sum_nk,  /* total instances of a word */
                as_index,   /* index of word in ea_word array - 0 indicates
			       non-existence */
                as_msg_freq; /* frequency of word in input document */
extendable_array ea_word,      /* a list of the unique words in the database */
                 ea_msg_words, /* set of unique words in message */
                 ea_category,  /* a list of the unique categories/maiboxes */
                 ea_lines,     /* the message broken into lines */
                 ea_orig_message; /* list of messages to be refiled */
int num_categories;   /* total number of distinct categories */
int num_words;        /* total number of distinct words */
int num_docs;         /* total number of documents */
int num_msg_words;    /* total number of distinct words in message */

/* Global variables for keeping track of time/speed of ifile */
clock_t DMZ_start, DMZ_end;

/* Global variable to keep track of mh_profile settings */
mh_profile_struct mh_profile;

char * dest_folder = NULL;  /* name of destination foler */
                            /* used for setting as_n when reading message */

/* irefile specific function prototypes */
void irefile_print_usage (void);
void irefile_read_idata (FILE * DATA);
int irefile_read_word_frequencies (FILE * DATA);
int irefile_process_frequency_line (char *line);


/* Main program */
int
main(int argc, char **argv)
{
  char * home_dir = NULL;       /* user's home directory, no ending '/' */
  char * source_folder = NULL;  /* name of source folder */
  char * source_file = NULL;
  char * msg_name = NULL;
  FILE * CONTEXT;               /* pointer for context file (cur folder) */
  char line[MAX_STR_LEN];       /* buffer for reading lines from a file */
  char * token = NULL;          /* stores current token when lexing text */
  FILE * MH_SEQ;                /* pointer for mh_sequences file (cur msg) */
  char * mh_seq = NULL;         /* full path of mh_sequences file */
  char * acc_file = NULL;       /* full path of accuracy file */
  char * source_folder_path = NULL; /* full path of source folder */
  char * msg_path = NULL;       /* full path of a message to be refiled */
  FILE * MSG = NULL;            /* pointer for reading message */
  char * data_file = NULL;      /* full path of idata file */
  FILE * DATA = NULL;           /* pointer to idata file */
  char * folders_file = NULL;   /* full path of listing of all mailboxes */

  int i, j;
  int skip_msg = 0;
  mh_profile_struct mh_profile;
  int words_written = 0;        // number of words written to .idata file

  /* Why classes & C++ can be nice:  (automatic initialization) */

  EXT_ARRAY_INIT(ea_word, char *, 3000);
  EXT_ARRAY_INIT(ea_msg_words, char *, 300);
  EXT_ARRAY_INIT(ea_category, char *, 50);
  EXT_ARRAY_INIT(ea_lines, char *, 100);
  EXT_ARRAY_INIT(ea_orig_message, char *, 5);
  assoc_array_int_init(as_n);
  assoc_array_int_init(as_docs);
  assoc_array_int_init(as_age);
  assoc_array_int_init(as_nk);
  assoc_array_int_init(as_sum_nk);
  assoc_array_int_init(as_index);
  assoc_array_int_init(as_msg_freq);

  ifile_open_log(argc, argv);

  /* Get home directory */
  home_dir = getenv("HOME");
  if (home_dir == NULL)
    ifile_error("Fatal: HOME environment variable not defined!\n");
  ifile_verbosify(ifile_verbose, "home directory = %s\n", home_dir);

  /* Reads in info from ~/.mh_profile */
  mh_profile = ifile_read_profile(home_dir);
  
  ifile_verbosify(ifile_progress, "Reading command line arguments...\n");

  /* variables set by getting command line args:

     source_folder
     dest_folder
     ea_orig_message
     */

  /* gets command line arguments */
  for (i=1; i < argc; i++)
    {
      if (strcmp(argv[i], "-src") == 0)
	{
	  if (i+1 >= argc)
	    {
	      ifile_verbosify(ifile_silent, "Expecting folder name after '-src' option.\n");
	      irefile_print_usage();
	    } 
	  else if (argv[i+1][0] == '+')
	    {
	      /* skip '+' when creating SOURCE_FOLDER */
	      source_folder = ifile_sprintf("%s", argv[i+1] + 1);
	      i++;
	      ifile_verbosify(ifile_verbose, "source folder = %s\n", source_folder);
	    }
	  else
	    {
	      ifile_verbosify(ifile_silent, "Error: folder name %s lacks preceeding \'+\' or is in wrong format.\n", source_folder);
	      irefile_print_usage();
	    }
	}
      else if (strcmp(argv[i], "-file") == 0)
	{
	  source_file = ifile_sprintf("%s", argv[i+1]);
	  ifile_verbosify(ifile_verbose, "source file = %s\n", source_file);
	  i++;
	}
      else if ((strcmp(argv[i], "--help") == 0) ||
	       (strcmp(argv[i], "-h") == 0))
	irefile_print_usage();
      else if ((strcmp(argv[i], "-draft") == 0) ||	    
	       (strcmp(argv[i], "-nolink") == 0) ||
	       (strcmp(argv[i], "-preserve") == 0) ||
	       (strcmp(argv[i], "-nopreserve") == 0) ||
	       (strcmp(argv[i], "-rmmproc") == 0))
	ifile_verbosify(ifile_debug, "%s recognized as MH command line argument\n", argv[i]);
      else if (strcmp(argv[i], "-link") == 0)
	{
	  ifile_close_log();
	  ifile_mh_refile(argc, argv);   /* Normally does not exit */
	  ifile_error("Not able to call MH refile.\nDon't know what to do with -link option!\nNo refile action taken.\n");
	}
      else if (argv[i][0] == '+')
	{
	  dest_folder = ifile_sprintf("%s", argv[i]+1);
	  ifile_verbosify(ifile_verbose, "destination folder = %s\n", dest_folder);
	}
      else
	{
	  /* this should take special care to handle the message named 'cur' */

	  char * msg_name = ifile_sprintf("%s", argv[i]);
	  j = ea_orig_message.last_elem;
	  EXT_ARRAY_SET(ea_orig_message, char *, j+1, msg_name);
	  ifile_verbosify(ifile_verbose, "ea_orig_message[%d] = %s\n",
			  j+1, msg_name);
	}
    }
  
  if (dest_folder == NULL)
    {
      ifile_verbosify(ifile_silent, "Error: No destination folder specified!\n");
      irefile_print_usage();
    }
  
  if (source_folder == NULL)
    {
      if ((CONTEXT = fopen(mh_profile.context_file, "r")) == NULL)
	{
	  ifile_verbosify(ifile_silent, "No source folder given.  No context file found: %s!\nYou must specify a source folder!\n", mh_profile.context_file);
	  irefile_print_usage();
	}
      
      readline(line, sizeof(line), CONTEXT);
      while (!feof(CONTEXT))
	{
	  token = strtok(line, " \t");
	  if (strcmp(token, "Current-Folder:") == 0)
	    {
	      token = strtok(NULL, " \t");
	      source_folder = ifile_sprintf("%s", token);
	      ifile_verbosify(ifile_verbose, "current MH folder = %s\n",
			      token);
	    }
	  readline(line, sizeof(line), CONTEXT);
	}
    }
  
  if (ea_orig_message.last_elem == -1)
    {
      mh_seq = ifile_sprintf("%s/%s%s/.mh_sequences", home_dir,
				    mh_profile.mail_path, source_folder);
      if ((MH_SEQ = fopen(mh_seq, "r")) == NULL)
	{
	  ifile_verbosify(ifile_silent, "Could not open MH sequences file: %s.\nNot able to determine current message.\nPlease specify a message to be refiled\n", mh_seq);
	  irefile_print_usage();
	}
      
      readline(line, sizeof(line), MH_SEQ);
      while (!feof(MH_SEQ))
	{
	  token = strtok(line, " \t");
	  if (strcmp(token, "cur:") == 0)
	    {
	      token = strtok(NULL, " \t");
	      msg_name = strdup(token);
	      EXT_ARRAY_SET(ea_orig_message, char *, 0, msg_name);
	      ifile_verbosify(ifile_verbose, "Current message is %s\n",
			      msg_name);
	    }
	  readline(line, sizeof(line), MH_SEQ);
	}
     }
  

  // if the destination folder has a .skip_me file, then the user does not
  // want mail filtered into that folder.  Thus, we should not record the fact
  // that this piece of mail is being moved there.  i.e. abort execution and
  // call the MH refile program

  if (fopen(ifile_sprintf("%s/%s%s/.skip_me", home_dir, mh_profile.mail_path,
			  dest_folder), "r"))
    ifile_mh_refile(argc, argv);

  
  if (ea_orig_message.last_elem == -1)
    {
      ifile_verbosify(ifile_verbose,
		      "No cur message, no source message(s) specified\n");
      irefile_print_usage();
    }

  if (source_file != NULL)
    {
      if (ea_orig_message.last_elem >= 0)
	{
	  ifile_verbosify(ifile_quiet, "Warning: Messages other than %s are being ignored.\n", source_file);
	  
	  /* no use for ea_orig_message strings - free used memory */
	  for (i=0; i <= ea_orig_message.last_elem; i++)
	    {
	      ifile_free(EXT_ARRAY_GET(ea_orig_message, char *, i));
	      EXT_ARRAY_SET(ea_orig_message, char *, i, NULL);
	    }
	  ea_orig_message.elems = 0;
	  ea_orig_message.last_elem = -1;
	}
      
      if (source_folder != NULL)
	{
	  ifile_verbosify(ifile_quiet, "Warning: \'-src\' command line option is being ignored\n");
	  ifile_free(source_folder);
	}
      /* set 0th ea_orig_message equal to source_file */
      msg_name = strdup(source_file);
      EXT_ARRAY_SET(ea_orig_message, char *, 0, msg_name);
    }

  acc_file = ifile_sprintf("%s/%s", home_dir, IFILE_ACCURACY);
  ifile_update_accuracy(acc_file, 0, ea_orig_message.elems);
  
  /* Now read the database */
  num_words = 0;
  num_categories = 0;
  
  data_file = ifile_sprintf("%s/.idata", home_dir);
  if ((DATA = fopen(data_file, "r")) != NULL)
    {
      ifile_verbosify(ifile_progress, "Opening %s for reading...\n",
		      data_file);
      irefile_read_idata(DATA);
    }
  else
    {
      ifile_verbosify(ifile_progress, "Not able to open %s for reading.\n",
		      data_file);
    }

  source_folder_path = ifile_sprintf("%s/%s%s/", home_dir,
				     mh_profile.mail_path, source_folder);
  ifile_verbosify(ifile_verbose, "source folder path = %s\n", source_folder_path);
  
  ifile_verbosify(ifile_progress, "Reading messages...\n");

  /* refiles messages one at a time */
  
  for (i=0; i <= ea_orig_message.last_elem; i++)
    {
      skip_msg = FALSE;
      
      if (source_file == NULL)
	{
	  msg_name = ifile_sprintf("%s",
				   EXT_ARRAY_GET(ea_orig_message, char *, i));
	  msg_path = ifile_sprintf("%s%s", source_folder_path, msg_name);
	}
      else
	msg_path = strdup(source_file);
      
      ifile_verbosify(ifile_verbose, "Reading message: %s\n", msg_path);
      
      if (source_folder != NULL)
	{
	  if ((MSG = fopen(msg_path, "r")) == NULL)
	    {
	      ifile_verbosify(ifile_quiet, "Warning: Could not open %s\n",
			      msg_path);
	      skip_msg = TRUE;
	    }
	}
      else /* the entire message path was specified on the command line */
	{
	  if ((MSG = fopen(msg_name, "r")) == NULL)
	    {
	      ifile_verbosify(ifile_quiet, "Warning: Could not open %s\n", 
			      msg_name);
	      skip_msg = TRUE;
	    }
	}
      
      /* clear out char *'s allocated during read of last message */
      EXT_ARRAY_FREE_ELEMS(ea_lines, char *);
      
      ifile_read_message(MSG);
      
      fclose(MSG);
      
      assoc_array_int_set(as_docs, dest_folder, 1 +
			  assoc_array_int_get(as_docs, dest_folder));	
      for (j=1; j <= ea_word.last_elem; j++)
	assoc_array_int_set(as_age, EXT_ARRAY_GET(ea_word, char *, j), 1 +
			    assoc_array_int_get(as_age, EXT_ARRAY_GET(ea_word, char *, j)));
    }

  // synchronize .idata file with list of MH folders.
  folders_file = ifile_sprintf("%s/%s.folders", home_dir,
			       mh_profile.mail_path);  
  if (ifile_update_categories(folders_file))
    ifile_verbosify(ifile_quiet, "Not able to read %s.\nContinuing to write .idata file with old list of folder names\n", folders_file);

  /* Reopen and write the learned information */
  
  data_file = ifile_sprintf("%s/.idata", home_dir);
  words_written = ifile_write_idata(data_file);
  if (words_written >= 0)
    fclose(DATA);
  
  DMZ_end = clock();
  ifile_verbosify(ifile_progress, "Wrote %d categories, %d words.  Elapsed time: %.3f sec\n", num_categories, words_written,
		  ((float)(DMZ_end-DMZ_start))/CLOCKS_PER_SECOND);
  
  /* if refile.bak is executable, then use it to refile all the messages */
  if (ifile_exe(MH_REFILE))
    ifile_mh_refile(argc, argv);

  ifile_error("Error: Not able to properly refile messages with MH refile!\n");
  
  return -1;
}


/* Function expects a file pointer which points to the beginning of the idata
   file.  Function will store information from data file into the global
   variables ea_category, as_n, num_categories, as_docs, tot_docs, ea_word,
   as_index, as_age and as_nk. 
   */
/* written by Jason Rennie <jr6b+@andrew.cmu.edu> and others for ifile */
void 
irefile_read_idata (FILE * DATA)
{
  int words_read, categories_read;
  DMZ_start=clock();
  
  categories_read = ifile_read_category_names (DATA);
  if (categories_read < 0) return;
  ifile_read_category_word_counts (DATA);
  ifile_read_category_doc_counts (DATA);
  words_read = irefile_read_word_frequencies (DATA);
  
  DMZ_end = clock();
  ifile_verbosify(ifile_progress,
		  "Read %d categories, %d words.  Elapsed time: %.3f sec\n",
		  categories_read, words_read,
	          ((float)(DMZ_end-DMZ_start))/CLOCKS_PER_SECOND);
  
}


/* Expects a valid file pointer which points to the beginning of the word
   frequencies section of an idata file.  Function will read lines and store
   the gleamed information into global variables. */
/* written by Jason Rennie <jr6b+@andrew.cmu.edu> and others for ifile */
int
irefile_read_word_frequencies (FILE * DATA)
{
  char line[MAX_STR_LEN*4];
  int words_read = 0;

  num_words = 0;

  readline(line, sizeof(line), DATA);
  while (!feof(DATA))
    {
      num_words++;
      words_read += irefile_process_frequency_line(line);
      readline(line, sizeof(line), DATA);
    }
  return words_read;
}


/* Given a line from an idata file, function determines whether information
   is needed and if so, parses line and stores information in global array.
   Currently parses all lines given to it. */
/* written by Jason Rennie <jr6b+@andrew.cmu.edu> and others for ifile */
int
irefile_process_frequency_line (char *line)
{
  int cat_num, freq, old_freq;
  char cur_word[MAX_STR_LEN];
  char * token = NULL;
  char freq_entry[MAX_STR_LEN];

  token = strtok(line, " \t");
  
  if (token == NULL)
    return 0;
  
  //  if (assoc_array_int_get(as_words, token) == 0)
  //    return;

  EXT_ARRAY_SET(ea_word, char *, num_words, strdup(token));
  assoc_array_int_set(as_index, token, num_words); 
  strcpy(cur_word, token);
  
  /* Read age of word */
  token = strtok(NULL, " \t");
  assoc_array_int_set(as_age, cur_word,	atoi(token)); 
  
  /* Read frequency per category cat_num:freq */
  token = strtok(NULL, ":");
  while (token != NULL)
    {	
      cat_num = atoi(token);
      token = strtok(NULL, " \t");	   
      freq = atoi(token);
      sprintf(freq_entry, "%s,%s", 
	      EXT_ARRAY_GET(ea_word, char*, num_words),
	      EXT_ARRAY_GET(ea_category, char*, cat_num));
      ifile_verbosify(ifile_debug, "%s = %d  ", freq_entry, freq);
      assoc_array_int_set(as_nk, freq_entry, freq);
      /* sums total frequency per word */
      old_freq = assoc_array_int_get(as_sum_nk, cur_word);
      assoc_array_int_set(as_sum_nk, cur_word, old_freq + freq);
      token = strtok(NULL, ":");
    }
  ifile_verbosify(ifile_debug, "\n");

  return 1;
}


/* Prints information about program usage and aborts execution */
/* Written by Jason Rennie <jr6b+@andrew.cmu.edu> for ifile */
void 
irefile_print_usage (void)
{
  ifile_close_log();

  fprintf(stderr, "Usage: irefile [msgs] [-src +folder] [-file file] [MH options] +folder\n");
  fprintf(stderr, "    or irefile --help\n\n");
  fprintf(stderr, " msgs         => text files to be refiled into FOLDER\n");
  fprintf(stderr, " -src +folder => looks for MSGS in specified folder\n");
  fprintf(stderr, " -file file   => disregards previous two options; uses 'file' as\n");
  fprintf(stderr, "                 source message\n");
  fprintf(stderr, " +folder      => folder to direct messages into\n");
  fprintf(stderr, " MH options   => other command line options will be passed to\n");
  fprintf(stderr, "                 MH refile, so all MH refile options are valid\n\n");
  fprintf(stderr, "ifile v%d.%d.%d\n", IFILE_MAJOR_VERSION_NUMBER,
          IFILE_MINOR_VERSION_NUMBER, IFILE_TRIFLING_VERSION_NUMBER);
  exit(0);
}
